import sys

PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] == 3

if PY2:
    textclass = unicode
    from HTMLParser import HTMLParser
    from itertools import izip_longest
    from StringIO import StringIO

else:
    textclass = str
    from html.parser import HTMLParser
    from io import StringIO
    from itertools import zip_longest as izip_longest



import contextlib
import shutil
import sys
import tempfile


def build_po_string(data):
    return (
        '#, fuzzy\n'
        'msgid ""\n'
        'msgstr ""\n'
        '"Project-Id-Version: foo\\n"\n'
        '"POT-Creation-Date: 2013-06-05 14:16-0700\\n"\n'
        '"PO-Revision-Date: 2010-04-26 18:00-0700\\n"\n'
        '"Last-Translator: Automatically generated\\n"\n'
        '"Language-Team: English\\n"\n'
        '"Language: \\n"\n'
        '"MIME-Version: 1.0\\n"\n'
        '"Content-Type: text/plain; charset=UTF-8\\n"\n'
        '"Content-Transfer-Encoding: 8bit\\n"\n'
        '"X-Generator: Translate Toolkit 1.6.0\\n"\n\n'
        + data)


@contextlib.contextmanager
def tempdir():
    """Builds a tempdir and cleans up afterwards

    Usage::

        with tempdir() as dir_:
            # blah blah blah

    """
    dir_ = tempfile.mkdtemp()
    yield dir_
    shutil.rmtree(dir_)


@contextlib.contextmanager
def redirect(stdin=None, stdout=None, stderr=None):
    """Redirects stdin, stdout and stderr

    Usage::

        stdout = StringIO()
        stderr = StringIO()

        with redirect(stdout=stdout, stderr=stderr):
            print 'blah blah'


        # do soething with stdout.getvalue()

    """
    old_stdin = sys.stdin
    if stdin:
        sys.stdin = stdin
    old_stdout = sys.stdout
    if stdout:
        sys.stdout = stdout
    old_stderr = sys.stderr
    if stderr:
        sys.stderr = stderr

    yield

    sys.stdin = old_stdin
    sys.stdout = old_stdout
    sys.stderr = old_stderr



from collections import namedtuple

import polib

from dennis.tools import VariableTokenizer, all_subclasses, parse_dennis_note


IdString = namedtuple('IdString', ('msgid_fields', 'msgid_strings'))


class LintedEntry(object):
    def __init__(self, poentry):
        self.poentry = poentry
        self.msgid = poentry.msgid

        if poentry.msgid_plural:
            msgid_fields = ('msgid', 'msgid_plural')
            msgid_strings = (poentry.msgid, poentry.msgid_plural)
        else:
            msgid_fields = ('msgid',)
            msgid_strings = (poentry.msgid,)

        self.str = IdString(msgid_fields, msgid_strings)

        self.warnings = []

    def add_warning(self, code, idstr, msg):
        self.warnings.append((code, idstr, msg))

    def has_problems(self):
        return bool(self.warnings)


class TemplateLintRule(object):
    num = ''
    name = ''
    desc = ''

    def lint(self, vartok, linted_entry):
        """Takes a linted entry and adds errors and warnings

        :arg vartok: the variable tokenizer to use for tokenizing
            on variable tokens
        :arg linted_entry: the LintedEntry to work on

        """
        raise NotImplemented


class HardToReadNamesTLR(TemplateLintRule):
    num = 'W500'
    name = 'hardtoread'
    desc = 'Looks for variables that are hard to read like o, O, 0, l, 1'

    hard_to_read = ('o', 'O', '0', 'l', '1')

    def lint(self, vartok, linted_entry):
        idstr = linted_entry.str

        for s in idstr.msgid_strings:
            if not s:
                continue

            msgid_tokens = vartok.extract_tokens(s)
            msgid_tokens = [vartok.extract_variable_name(token)
                            for token in msgid_tokens]

            for token in msgid_tokens:
                if token in self.hard_to_read:
                    linted_entry.add_warning(
                        self.num,
                        idstr,
                        u'hard to read variable name "{0}"'.format(
                            token))


class OneCharNamesTLR(TemplateLintRule):
    num = 'W501'
    name = 'onechar'
    desc = 'Looks for one character variable names'

    def lint(self, vartok, linted_entry):
        idstr = linted_entry.str

        for s in idstr.msgid_strings:
            if not s:
                continue

            msgid_tokens = vartok.extract_tokens(s)
            msgid_tokens = [vartok.extract_variable_name(token)
                            for token in msgid_tokens]

            for token in msgid_tokens:
                if len(token) == 1 and token.isalpha():
                    linted_entry.add_warning(
                        self.num,
                        idstr,
                        u'one character variable name "{0}"'.format(
                            token))


class MultipleUnnamedVarsTLR(TemplateLintRule):
    num = 'W502'
    name = 'unnamed'
    desc = 'Looks for multiple unnamed variables'

    def lint(self, vartok, linted_entry):
        idstr = linted_entry.str

        for s in idstr.msgid_strings:
            if not s:
                continue

            msgid_tokens = vartok.extract_tokens(s, unique=False)
            msgid_tokens = [vartok.extract_variable_name(token)
                            for token in msgid_tokens]

            if msgid_tokens.count('') > 1:
                linted_entry.add_warning(
                    self.num,
                    idstr,
                    u'multiple variables with no name.')


def get_available_lint_rules(name_and_num=False):
    lint_rules = {}

    for cls in all_subclasses(TemplateLintRule):
        if cls.num:
            lint_rules[cls.num] = cls
            if name_and_num and cls.name:
                lint_rules[cls.name] = cls

    return lint_rules


class InvalidRulesSpec(Exception):
    pass


def convert_rules(rules_spec):
    # This removes empty strings from the rules_spec.
    rules_spec = [rule for rule in rules_spec if rule]

    if not rules_spec:
        lint_rules = get_available_lint_rules()
        return [rule() for num, rule in lint_rules.items()]

    try:
        lint_rules = get_available_lint_rules(name_and_num=True)
        rules = [lint_rules[rule]() for rule in rules_spec]
    except KeyError:
        raise InvalidRulesSpec(rules_spec)

    return rules


class TemplateLinter(object):
    def __init__(self, vars_, rules_spec):
        self.vartok = VariableTokenizer(vars_)
        self.rules_spec = rules_spec
        self.rules = convert_rules(self.rules_spec)

    def lint_poentry(self, poentry):
        linted_entry = LintedEntry(poentry)

        skip = parse_dennis_note(poentry.comment)

        # Check the comment to see if what we should ignore.
        for lint_rule in self.rules:
            if skip == '*' or lint_rule.num in skip:
                continue

            lint_rule.lint(self.vartok, linted_entry)

        return linted_entry

    def verify_file(self, filename_or_string):
        """Verifies strings in file.

        :arg filename_or_string: filename to verify or the contents of
            a pofile as a string

        :returns: list of LintedEntry objects each with errors and
            warnings

        :raises IOError: if the file is not a valid .po file or
            doesn't exist
        """
        po = polib.pofile(filename_or_string)
        return [
            self.lint_poentry(entry) for entry in po
        ]



import os
from unittest import TestCase

from nose.tools import eq_
import polib

from dennis.cmdline import linttemplate_cmd
from dennis.minisix import StringIO
from dennis.templatelinter import (
    HardToReadNamesTLR,
    OneCharNamesTLR,
    MultipleUnnamedVarsTLR,
    TemplateLinter,
    LintedEntry,
)
from dennis.tools import VariableTokenizer
from dennis.tests import build_po_string, redirect, tempdir


class LinTemplateCmdTest(TestCase):
    def test_empty(self):
        stdout = StringIO()
        with redirect(stdout=stdout):
            linttemplate_cmd('test', 'linttemplate', [])

    def test_basic(self):
        with tempdir() as dir_:
            pofile = build_po_string(
                '#: foo/foo.py:5\n'
                'msgid "Foo: %(l)s"\n'
                'msgstr ""\n')

            fn = os.path.join(dir_, 'messages.pot')
            with open(fn, 'w') as fp:
                fp.write(pofile)

            stdout = StringIO()
            stderr = StringIO()
            with redirect(stdout=stdout, stderr=stderr):
                linttemplate_cmd('test', 'linttemplate', [fn])

            assert '>>> Working on:' in stdout.getvalue()

    def test_no_files_to_work_on(self):
        stdout = StringIO()
        stderr = StringIO()
        with redirect(stdout=stdout, stderr=stderr):
            ret = linttemplate_cmd('test', 'linttemplate', ['foo'])

        eq_(ret, 1)
        assert 'No files to work on.' in stderr.getvalue()

    def test_file_not_exists(self):
        with tempdir() as dir_:
            fn = os.path.join(dir_, 'messages.pot')
            stdout = StringIO()
            stderr = StringIO()
            with redirect(stdout=stdout, stderr=stderr):
                ret = linttemplate_cmd('test', 'linttemplate', [fn])

            eq_(ret, 1)
            assert (
                'IOError' in stderr.getvalue()
                or 'OSError' in stderr.getvalue()
            )
            assert 'does not exist' in stderr.getvalue()


class TemplateLinterTest(TestCase):
    def test_linter(self):
        pofile = build_po_string(
            '#: foo/foo.py:5\n'
            'msgid "Foo"\n'
            'msgstr ""\n')

        linter = TemplateLinter(['pysprintf', 'pyformat'], ['W500'])
        results = linter.verify_file(pofile)

        # This should give us one linted entry with no errors
        # and no warnings in it.
        eq_(len(results), 1)
        eq_(len(results[0].warnings), 0)


def build_linted_entry(po_data):
    po = polib.pofile(build_po_string(po_data))
    poentry = list(po)[0]
    return LintedEntry(poentry)


class TLRTestCase(TestCase):
    vartok = VariableTokenizer(['pysprintf', 'pyformat'])


class HardToReadNamesTLRTestCase(TLRTestCase):
    lintrule = HardToReadNamesTLR()

    def test_hard_to_read_names(self):
        for c in ('o', 'O', '0', 'l', '1'):
            linted_entry = build_linted_entry(
                '#: foo/foo.py:5\n'
                'msgid "Foo: %(' + c + ')s"\n'
                'msgstr ""\n')

            self.lintrule.lint(self.vartok, linted_entry)

            eq_(len(linted_entry.warnings), 1)

            linted_entry = build_linted_entry(
                '#: foo/foo.py:5\n'
                'msgid "Foo: {' + c + '}"\n'
                'msgstr ""\n')

            self.lintrule.lint(self.vartok, linted_entry)

            eq_(len(linted_entry.warnings), 1)


class MultipleUnnamedVarsTLRTestCase(TLRTestCase):
    lintrule = MultipleUnnamedVarsTLR()

    def test_multi_vars_no_name(self):
        linted_entry = build_linted_entry(
            '#: foo/foo.py:5\n'
            'msgid "Foo: %s %s"\n'
            'msgstr ""\n')

        self.lintrule.lint(self.vartok, linted_entry)

        eq_(len(linted_entry.warnings), 1)

        linted_entry = build_linted_entry(
            '#: foo/foo.py:5\n'
            'msgid "Foo: {} {}"\n'
            'msgstr ""\n')

        self.lintrule.lint(self.vartok, linted_entry)

        eq_(len(linted_entry.warnings), 1)


class OneCharNamesTLRTestCase(TLRTestCase):
    lintrule = OneCharNamesTLR()

    def test_one_character_names(self):
        linted_entry = build_linted_entry(
            '#: foo/foo.py:5\n'
            'msgid "Foo: %(c)s"\n'
            'msgstr ""\n')

        self.lintrule.lint(self.vartok, linted_entry)

        eq_(len(linted_entry.warnings), 1)

        linted_entry = build_linted_entry(
            '#: foo/foo.py:5\n'
            'msgid "Foo: {c}"\n'
            'msgstr ""\n')

        self.lintrule.lint(self.vartok, linted_entry)

        eq_(len(linted_entry.warnings), 1)
